@@ -712,17 +712,47 @@ html_view_leave_notify (GtkWidget *widget, GdkEventCrossing *event)
 	return TRUE;
 }
 
+static gboolean
+redraw_idle_callback (GtkWidget *widget)
+{
+	gtk_widget_queue_draw (widget);
+	return FALSE;
+}
+
 static gboolean
 html_view_expose (GtkWidget *widget, GdkEventExpose *event)
 {
 	HtmlView *view;
 	gboolean retval;
+	GdkRectangle area;
 
 	view = HTML_VIEW (widget);
 
 	if (!GTK_WIDGET_DRAWABLE (widget) || (event->window != GTK_LAYOUT (view)->bin_window))
 		return FALSE;
 
+	if (GTK_WIDGET_HAS_FOCUS (widget) &&
+	    !view->document->focus_element) {
+		GtkLayout *layout;
+
+		layout = GTK_LAYOUT (view);
+		area.x = (gint) gtk_adjustment_get_value (layout->hadjustment);
+		area.y = (gint) gtk_adjustment_get_value (layout->vadjustment);
+		area.width = widget->allocation.width;
+		area.height = widget->allocation.height;
+
+		if (event->area.x != area.x ||
+		    event->area.y != area.y ||
+		    event->area.width != area.width ||
+		    event->area.height != area.height) {
+			/*
+			 * We need to invalidate the entire area so that the
+			 * focus area is redrawn
+			 */
+			gtk_idle_add (redraw_idle_callback, widget);
+			return TRUE;
+		}
+	}
 	html_view_paint (view, &event->area);
 	retval = GTK_WIDGET_CLASS (parent_class)->expose_event (widget, event);
 	if (GTK_WIDGET_HAS_FOCUS (widget) &&
@@ -738,8 +768,9 @@ html_view_expose (GtkWidget *widget, GdkEventExpose *event)
 				      "focus-line-width", &focus_width,
 				      "focus-padding", &focus_pad,
 				      NULL);
-		x = focus_pad;
-		y = focus_pad;
+		x = area.x + focus_pad;
+		y = area.y + focus_pad;
+
 		width = widget->allocation.width - (focus_width + 2 * focus_pad);
 		height = widget->allocation.height - (focus_width + 2 * focus_pad);
 		gtk_paint_focus (widget->style, GTK_LAYOUT (view)->bin_window,
@@ -840,19 +871,45 @@ focus_element_destroyed (gpointer data)
 	g_object_set_data (G_OBJECT (data), "saved-focus", NULL);
 }
 
+/* We must release all references to other objects here. But be careful as
+   it can be called multiple times. */
 static void
-html_view_finalize (GObject *object)
+html_view_destroy (GtkObject *object)
 {
 	HtmlView *view = HTML_VIEW (object);
+
 	gpointer saved_focus;
 
+	if (view->relayout_timeout_id != 0) {
+		gtk_timeout_remove (view->relayout_timeout_id);
+		view->relayout_timeout_id = 0;
+	}
+
+	if (view->relayout_idle_id != 0) {
+		gtk_idle_remove (view->relayout_idle_id);
+		view->relayout_idle_id = 0;
+	}
+
 	saved_focus = g_object_get_data (G_OBJECT (view), "saved-focus");
 	if (saved_focus) {
 		g_object_weak_unref (G_OBJECT (saved_focus),
 			   	     (GWeakNotify) focus_element_destroyed,
 				     view);
+		g_object_set_data (G_OBJECT (view), "saved-focus", NULL);
 	}
-	g_object_unref (view->document);
+
+	if (view->document) {
+		g_object_unref (view->document);
+		view->document = NULL;
+	}
+
+	GTK_OBJECT_CLASS (parent_class)->destroy (object);
+}
+
+static void
+html_view_finalize (GObject *object)
+{
+	/*HtmlView *view = HTML_VIEW (object);*/
 
 	if (G_OBJECT_CLASS (parent_class)->finalize)
 		(G_OBJECT_CLASS (parent_class)->finalize) (object);
@@ -865,9 +922,6 @@ html_view_focus (GtkWidget *widget, GtkDirectionType direction)
 	HtmlBox *box;
 	DomElement *old_focus_element;
 
-	if (view->moving_focus_out)
-		return FALSE;
-
 	if (!GTK_WIDGET_HAS_FOCUS (widget)) {
 		gtk_widget_grab_focus (widget);
 		return TRUE;
@@ -958,20 +1012,12 @@ html_view_real_activate (HtmlView *view)
 
 static void html_view_real_move_focus_out (HtmlView *view, GtkDirectionType direction_type)
 {
-	GtkWidget *toplevel;
-
-	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (view));
-	if (!GTK_WIDGET_TOPLEVEL (toplevel))
-		return;
+	GtkWidget *widget = GTK_WIDGET (view);
 
 	html_document_update_focus_element (view->document, NULL);
-	g_object_ref (view);
-
-	view->moving_focus_out = TRUE;
-	g_signal_emit_by_name (toplevel, "move_focus", direction_type);
-	view->moving_focus_out = FALSE;
-
-	g_object_unref (view);
+	if (!GTK_WIDGET_HAS_FOCUS (widget))
+		gtk_widget_grab_focus (widget);
+	gtk_widget_queue_draw (widget);
 }
 
 static void
@@ -1009,7 +1055,7 @@ html_view_set_document (HtmlView *view, HtmlDocument *document)
 	
 	if (view->document != NULL) {
 		html_view_disconnect_document (view, view->document);
-		
+		g_object_unref (view->document);
 		html_view_layout_tree_free (view, view->root);
 	}
 	
@@ -1128,6 +1174,7 @@ static void
 html_view_class_init (HtmlViewClass *klass)
 {
 	GObjectClass *object_class = (GObjectClass *)klass;
+	GtkObjectClass *gtkobject_class = (GtkObjectClass *)klass;
 	GtkWidgetClass *widget_class = (GtkWidgetClass *)klass;
 	GtkContainerClass *container_class = (GtkContainerClass *)klass;
 	GtkBindingSet *binding_set;
@@ -1137,6 +1184,8 @@ html_view_class_init (HtmlViewClass *klass)
 
 	object_class->finalize = html_view_finalize;
 
+	gtkobject_class->destroy = html_view_destroy;
+
 	widget_class->focus = html_view_focus;
 	
 	widget_class->size_allocate = html_view_size_allocate;
@@ -1258,7 +1307,6 @@ html_view_init (HtmlView *view)
 	view->relayout_idle_id = 0;
 	view->relayout_timeout_id = 0;
 	view->magnification = 1.0;
-	view->moving_focus_out = FALSE;
 }
 
 GType
